"""Run molecular dynamics with learned potentials.

Calculators are available for ASE and LAMMPS, but can be
extended to support your favorite MD software.
"""

from .ase_calc import FrankenCalculator
from .lammps_calc import LammpsFrankenCalculator

__all__ = (
    "FrankenCalculator",
    "LammpsFrankenCalculator",
)



from functools import reduce
import operator
from pathlib import Path
import zipfile

from franken.datasets.registry import DATASET_REGISTRY, BaseRegisteredDataset
from franken.utils.file_utils import download_file

TM23_ELEMENTS = [
    "Ag",
    "Au",
    "Cd",
    "Co",
    "Cr",
    "Cu",
    "Fe",
    "Hf",
    "Hg",
    "Ir",
    "Mn",
    "Mo",
    "Nb",
    "Ni",
    "Os",
    "Pd",
    "Pt",
    "Re",
    "Rh",
    "Ru",
    "Ta",
    "Tc",
    "Ti",
    "V",
    "W",
    "Zn",
    "Zr",
]

TM23_DATASETS = list(
    reduce(
        operator.concat,
        [
            [f"TM23/{el}", f"TM23/{el}-cold", f"TM23/{el}-warm", f"TM23/{el}-melt"]
            for el in TM23_ELEMENTS
        ],
    )
)


@DATASET_REGISTRY.register(TM23_DATASETS)
class TM23RegisteredDataset(BaseRegisteredDataset):
    relative_paths = reduce(
        operator.ior,
        [
            {
                f"TM23/{el}": {
                    "train": f"TM23/{el}_2700cwm_train.xyz",
                    "val": f"TM23/{el}_2700cwm_test.xyz",
                },
                f"TM23/{el}-cold": {
                    "train": f"TM23/{el}_cold_nequip_train.xyz",
                    "val": f"TM23/{el}_cold_nequip_test.xyz",
                },
                f"TM23/{el}-warm": {
                    "train": f"TM23/{el}_warm_nequip_train.xyz",
                    "val": f"TM23/{el}_warm_nequip_test.xyz",
                },
                f"TM23/{el}-melt": {
                    "train": f"TM23/{el}_melt_nequip_train.xyz",
                    "val": f"TM23/{el}_melt_nequip_test.xyz",
                },
            }
            for el in TM23_ELEMENTS
        ],
        {},
    )  # merge list of dicts

    @classmethod
    def get_path(
        cls, name: str, split: str, base_path: Path | None, download: bool = True
    ):
        if base_path is None:
            raise KeyError(None)
        relative_path = cls.relative_paths[name][split]
        path = base_path / relative_path
        if not path.is_file() and download:
            cls.download(base_path)
        if path.is_file():
            return path
        else:
            raise ValueError(f"Dataset not found at '{path.resolve()}'")

    @classmethod
    def download(cls, base_path: Path):
        tm23_base_path = base_path / "TM23"
        tm23_base_path.mkdir(exist_ok=True, parents=True)
        # Download
        download_file(
            url="https://archive.materialscloud.org/record/file?record_id=2113&filename=benchmarking_master_collection-20240316T202423Z-001.zip",
            filename=tm23_base_path / "data.zip",
            desc="Downloading TM23 dataset",
        )
        # Extract
        with zipfile.ZipFile(tm23_base_path / "data.zip", "r") as zf:
            zf.extractall(tm23_base_path)
        # Move files up one level
        for origin in (tm23_base_path / "benchmarking_master_collection").glob("*"):
            origin.rename(tm23_base_path / origin.name)
        # Cleanup
        (tm23_base_path / "data.zip").unlink()
        (tm23_base_path / "benchmarking_master_collection").rmdir()



"""Train franken from data."""

from franken.trainers.base import BaseTrainer
from franken.trainers.rf_cuda_lowmem import RandomFeaturesTrainer

__all__ = (
    "BaseTrainer",
    "RandomFeaturesTrainer",
)



"""
Test the model conversion to LAMMPS (essentially testing torch-scriptability, not LAMMPS directly)
"""

import os
import pytest
import torch

from franken.config import GaussianRFConfig, MaceBackboneConfig, MultiscaleGaussianRFConfig
from franken.data import BaseAtomsDataset
from franken.rf.model import FrankenPotential
from franken.rf.scaler import Statistics
from franken.utils.misc import garbage_collection_cuda
from franken.datasets.registry import DATASET_REGISTRY
from franken.calculators.lammps_calc import LammpsFrankenCalculator

from .conftest import DEVICES
from .utils import are_dicts_close, cleanup_dir, create_temp_dir


RF_PARAMETRIZE = [
    GaussianRFConfig(num_random_features=128, length_scale=1.0),
    MultiscaleGaussianRFConfig(num_random_features=128),
]


@pytest.mark.parametrize("rf_cfg", RF_PARAMETRIZE)
@pytest.mark.parametrize("device", DEVICES)
def test_lammps_compile(rf_cfg, device):
    """Test for checking save and load methods of FrankenPotential"""
    gnn_cfg = MaceBackboneConfig("MACE-L0")
    temp_dir = None
    try:
        # Step 1: Create a temporary directory for saving the model
        temp_dir = create_temp_dir()

        data_path = DATASET_REGISTRY.get_path("test", "test", None, False)
        dataset = BaseAtomsDataset.from_path(
            data_path=data_path,
            split="train",
            gnn_config=gnn_cfg,
        )
        model = FrankenPotential(
            gnn_config=gnn_cfg,
            rf_config=rf_cfg,
            scale_by_Z=True,
            num_species=dataset.num_species,
        ).to(device)

        with torch.no_grad():
            gnn_features_stats = Statistics()
            for data, _ in dataset:  # type: ignore
                data = data.to(device=device)
                gnn_features = model.gnn.descriptors(data)
                gnn_features_stats.update(
                    gnn_features, atomic_numbers=data.atomic_numbers
                )

            model.input_scaler.set_from_statistics(gnn_features_stats)
            garbage_collection_cuda()

        # Step 2: Save the model to the temporary directory
        model_save_path = os.path.join(temp_dir, "model_checkpoint.pth")
        model.save(model_save_path)

        # Step 3: Run create_lammps_model
        comp_model_path = LammpsFrankenCalculator.create_lammps_model(model_path=model_save_path, rf_weight_id=None)

        # Step 4: Load saved model
        comp_model = torch.jit.load(comp_model_path, map_location=device)

        # Step 4: Compare rf.state_dict between the original and loaded models
        with pytest.raises(RuntimeError) as exc:
            assert are_dicts_close(
                model.rf.state_dict(), comp_model.model.rf.state_dict(), verbose=True
            )
        assert "Float did not match Double" in str(exc.value)
        assert are_dicts_close(
            model.rf.double().state_dict(), comp_model.model.rf.state_dict(), verbose=True
        ), "The rf.state_dict() of the loaded model does not match the original model."

        with pytest.raises(RuntimeError) as exc:
            assert are_dicts_close(
                model.input_scaler.state_dict(),
                comp_model.model.input_scaler.state_dict(),
                verbose=True,
            )
        assert "Float did not match Double" in str(exc.value)
        assert are_dicts_close(
            model.input_scaler.double().state_dict(),
            comp_model.model.input_scaler.state_dict(),
            verbose=True,
        ), "The input_scaler.state_dict() of the loaded model does not match the original model."

        with pytest.raises(RuntimeError) as exc:
            assert are_dicts_close(
                model.energy_shift.state_dict(),
                comp_model.model.energy_shift.state_dict(),
                verbose=True,
            )
        assert "Float did not match Double" in str(exc.value)
        assert are_dicts_close(
            model.energy_shift.double().state_dict(),
            comp_model.model.energy_shift.state_dict(),
            verbose=True,
        ), "The energy_shift.state_dict() of the loaded model does not match the original model."
    finally:
        if temp_dir is not None:
            cleanup_dir(temp_dir)




import argparse
import dataclasses
import types
import typing
import docstring_parser
from typing import Any, Sequence

from franken.config import (
    AutotuneConfig,
    MaceBackboneConfig,
    FairchemBackboneConfig,
    SevennBackboneConfig,
    GaussianRFConfig,
    MultiscaleGaussianRFConfig,
    SolverConfig,
    DatasetConfig,
    HPSearchConfig,
)


def parse_docstring_from_object(obj: object) -> dict[str, str]:
    return {
        doc.arg_name: doc.description
        for doc in docstring_parser.parse_from_object(obj).params
        if doc.description is not None
    }


def get_field_docstring(cls: typing.Type, field_name: str):
    """Taken from tyro https://github.com/brentyi/tyro/blob/main/src/tyro/_docstrings.py#L25"""
    # NoneType will break docstring_parser.
    if cls is type(None):
        return None

    # Try to parse using docstring_parser.
    for cls_search in cls.__mro__:
        if cls_search.__module__ == "builtins":
            continue  # Skip `object`, `Callable`, `tuple`, etc.
        docstring = parse_docstring_from_object(cls_search).get(field_name, None)
        if docstring is not None:
            return docstring.strip()


def field_is_optional(field_type: Any):
    """A field is optional when it has Union type with a NoneType alternative.
    Note that Optional[] is a special form which is converted to a Union with a NoneType option
    """
    origin = typing.get_origin(field_type)
    if origin is typing.Union:
        return type(None) in typing.get_args(field_type)
    if origin is types.UnionType:
        return type(None) in typing.get_args(field_type)
    return False


def field_is_optional_str(field_type: Any):
    if not field_is_optional(field_type):
        return False
    is_str_present = type(str) in typing.get_args(field_type)
    return is_str_present and len(typing.get_args(field_type)) == 2


def field_is_optional_literal(field_type: Any):
    if not field_is_optional(field_type):
        return False
    all_subtypes = typing.get_args(field_type)
    # To check for literal we must check the origin of the sub-types
    all_subtype_origins = [typing.get_origin(t) for t in all_subtypes]
    is_lit_present = typing.Literal in all_subtype_origins
    return is_lit_present and len(all_subtypes) == 2


def parse_optional_literal(s: str) -> str | None:
    if s == "none":
        return None
    return s


class Argument:
    def __init__(self, full_name: str, dataclass, dataclass_name: str, **kwargs):
        self.full_name = full_name
        self.dataclass = dataclass
        self.dataclass_name = dataclass_name
        self.help = get_field_docstring(dataclass, dataclass_name)
        self.arg_config = kwargs
        self.is_arg_required = "default" not in self.arg_config
        self.dest = None

    def add(self, arg_container, **kwargs):
        try:
            action = arg_container.add_argument(
                f"--{self.full_name}", help=self.help, **kwargs
            )
        except TypeError as e:
            raise RuntimeError(
                f"Failed to add argument {self} with arguments {kwargs}."
            ) from e
        self.dest = action.dest

    def is_added(self) -> bool:
        return self.dest is not None

    @staticmethod
    def from_dataclass(
        dataclass, dataclass_name, full_name, opposite_full_name=None, **kwargs
    ):
        all_fields = dataclasses.fields(dataclass)
        try:
            named_field = [f for f in all_fields if f.name == dataclass_name][0]
        except IndexError:
            raise ValueError(
                f"Dataclass {dataclass} does not have attribute {dataclass_name}"
            )
        arg_kwargs = {}

        if named_field.default != dataclasses.MISSING:
            arg_kwargs["default"] = str(named_field.default)
        elif named_field.default_factory != dataclasses.MISSING:
            arg_kwargs["default"] = str(named_field.default_factory())

        if named_field.type in {int, float, str}:
            arg_kwargs["type"] = named_field.type

        # Literals
        type_origin = typing.get_origin(named_field.type)
        if type_origin == typing.Literal:
            arg_kwargs["choices"] = typing.get_args(named_field.type)
        if field_is_optional_literal(named_field.type):
            arg_kwargs["choices"] = list(typing.get_args(named_field.type)) + ["none"]
            arg_kwargs["type"] = parse_optional_literal

        # Booleans
        if named_field.type is bool:
            if named_field.default is False:
                arg_kwargs["action"] = "store_true"
            elif named_field.default is True:
                arg_kwargs["action"] = "store_false"
                if opposite_full_name is None:
                    raise ValueError(
                        f"opposite_full_name must be non-null for field {dataclass_name}"
                    )
                full_name = opposite_full_name

        return Argument(
            full_name=full_name,
            dataclass=dataclass,
            dataclass_name=dataclass_name,
            **(arg_kwargs | kwargs),
        )

    def __repr__(self):
        return f"Argument({self.full_name=} {self.dataclass=} {self.dataclass_name=})"


class ArgumentGroup:
    def __init__(
        self,
        name: str,
        title: str,
        desc: str,
        data_class: type,
        arguments: Sequence[Argument],
    ):
        self.name = name
        self.title = title
        self.desc = desc
        self.data_class = data_class
        self.arguments = arguments

    def add_to_parser(self, parser: argparse.ArgumentParser):
        arg_group = parser.add_argument_group(
            title=self.title, description=self.desc if self.desc else None
        )
        for arg in self.arguments:
            arg.add(arg_group, **(arg.arg_config | {"required": arg.is_arg_required}))

    def to_dataclass(self, args):
        dc_args = {}
        for arg_config in self.arguments:
            if arg_config.dest is None:
                raise RuntimeError(
                    f"dest attribute for argument {arg_config.full_name} is "
                    f"None. Have you parsed arguments correctly?"
                )
            dc_args[arg_config.dataclass_name] = getattr(args, arg_config.dest)
        return self.data_class(**dc_args)


class MutuallyExclusiveArgumentGroup(ArgumentGroup):
    def __init__(self, name: str, title: str, desc: str, data_class: type, arguments):
        for arg in arguments:
            arg.full_name = f"{name}.{arg.full_name}"  # e.g. fairchem.interaction_block
        super().__init__(name, title, desc, data_class, arguments)

    def add_to_parser(self, parser: argparse.ArgumentParser):
        arg_group = parser.add_argument_group(
            title=self.title, description=self.desc if self.desc else None
        )
        for arg in self.arguments:
            arg.add(arg_group, **arg.arg_config)


class GroupWithIndividualOptions:
    def __init__(
        self,
        group_name: str,
        me_groups: Sequence[MutuallyExclusiveArgumentGroup],
        help_text: str,
    ):
        self.group_name = group_name
        self.me_groups = {g.name: g for g in me_groups}
        self.help_text = help_text

    def add_to_parser(self, parser):
        parser.add_argument(
            f"--{self.group_name}",
            choices=list(self.me_groups.keys()),
            required=True,
            help=self.help_text,
        )
        for g_name, g in self.me_groups.items():
            g.add_to_parser(parser)

    def get_selected_subgroup(self, args) -> tuple[MutuallyExclusiveArgumentGroup, str]:
        parsed_subgroup_name = getattr(args, self.group_name, None)
        if parsed_subgroup_name is None:
            raise ValueError(
                f"Group {self.group_name} does not have any value in parsed arguments"
            )
        try:
            selected_subgroup = self.me_groups[parsed_subgroup_name]
        except KeyError:
            raise ValueError(
                f"Group {self.group_name} with value {parsed_subgroup_name} has no registered group data."
            )
        return selected_subgroup, parsed_subgroup_name

    def validate_required(self, parser, args):
        selected_subgroup, subgroup_name = self.get_selected_subgroup(args)

        for arg_config in selected_subgroup.arguments:
            if arg_config.dest is None:
                raise RuntimeError(
                    f"dest attribute for argument {arg_config.full_name} is "
                    f"None. Have you parsed arguments correctly?"
                )
            if arg_config.is_arg_required and getattr(args, arg_config.dest) is None:
                parser.error(
                    f"--{arg_config.full_name} is required when --{self.group_name}={subgroup_name}"
                )

    def to_dataclass(self, args):
        selected_subgroup, _ = self.get_selected_subgroup(args)
        return selected_subgroup.to_dataclass(args)


def get_args_from_dataclass(dataclass, full_name_prefix=""):
    args = []
    for f in dataclasses.fields(dataclass):
        full_name = f"{full_name_prefix}.{f.name.replace('_', '-')}"
        args.append(Argument.from_dataclass(dataclass, f.name, full_name))


def get_arg_groups():
    solver_arg_group = ArgumentGroup(
        "solver",
        "Solver",
        desc="",
        data_class=SolverConfig,
        arguments=[
            Argument.from_dataclass(
                SolverConfig,
                "l2_penalty",
                "l2-penalty",
                metavar="HYPERPARAMETER",
                type=HPSearchConfig.from_str,
            ),
            Argument.from_dataclass(
                SolverConfig,
                "force_weight",
                "force-weight",
                metavar="HYPERPARAMETER",
                type=HPSearchConfig.from_str,
            ),
        ],
    )
    dset_arg_group = ArgumentGroup(
        "dataset",
        "Dataset",
        desc="",
        data_class=DatasetConfig,
        arguments=[
            Argument.from_dataclass(DatasetConfig, "name", "dataset-name"),
            Argument.from_dataclass(DatasetConfig, "train_path", "train-path"),
            Argument.from_dataclass(DatasetConfig, "test_path", "test-path"),
            Argument.from_dataclass(DatasetConfig, "val_path", "val-path"),
            Argument.from_dataclass(
                DatasetConfig, "max_train_samples", "max-train-samples"
            ),
        ],
    )
    bbone_groups = GroupWithIndividualOptions(
        "backbone",
        [
            MutuallyExclusiveArgumentGroup(
                "mace",
                title="MACE backbone",
                desc="Configure the MACE backbone. Specify ``--backbone=mace`` to enable.",
                data_class=MaceBackboneConfig,
                arguments=[
                    Argument.from_dataclass(
                        MaceBackboneConfig, "path_or_id", "path-or-id"
                    ),
                    Argument.from_dataclass(
                        MaceBackboneConfig, "interaction_block", "interaction-block"
                    ),
                ],
            ),
            MutuallyExclusiveArgumentGroup(
                "sevenn",
                title="SevenNet backbone",
                desc="Configure the MACE backbone. Specify ``--backbone=sevenn`` to enable.",
                data_class=SevennBackboneConfig,
                arguments=[
                    Argument.from_dataclass(
                        SevennBackboneConfig, "path_or_id", "path-or-id"
                    ),
                    Argument.from_dataclass(
                        SevennBackboneConfig, "interaction_block", "interaction-block"
                    ),
                    Argument.from_dataclass(
                        SevennBackboneConfig,
                        "extract_after_act",
                        "extract-after-act",
                        "extract-before-act",
                    ),
                    Argument.from_dataclass(
                        SevennBackboneConfig,
                        "append_layers",
                        "append-layers",
                        "last-layer-only",
                    ),
                ],
            ),
            MutuallyExclusiveArgumentGroup(
                "fairchem",
                title="Fairchem backbone",
                desc="Configure the MACE backbone. Specify ``--backbone=fairchem`` to enable.",
                data_class=FairchemBackboneConfig,
                arguments=[
                    Argument.from_dataclass(
                        FairchemBackboneConfig, "path_or_id", "path-or-id"
                    ),
                    Argument.from_dataclass(
                        FairchemBackboneConfig, "interaction_block", "interaction-block"
                    ),
                ],
            ),
        ],
        help_text="The GNN backbone which will be used by franken.",
    )
    rf_groups = GroupWithIndividualOptions(
        "rf",
        [
            MutuallyExclusiveArgumentGroup(
                "gaussian",
                title="Gaussian RFs",
                desc="Configuration group for orthogonal random fourier features (ORFF), which are used to approximate a Gaussian kernel. Specify ``--rf=gaussian`` to enable this RF approximation",
                data_class=GaussianRFConfig,
                arguments=[
                    Argument.from_dataclass(
                        GaussianRFConfig, "num_random_features", "num-rf"
                    ),
                    Argument.from_dataclass(
                        GaussianRFConfig,
                        "length_scale",
                        "length-scale",
                        metavar="HYPERPARAMETER",
                        type=HPSearchConfig.from_str,
                    ),
                    Argument.from_dataclass(
                        GaussianRFConfig, "use_offset", "use-offset", "no-use-offset"
                    ),
                    Argument.from_dataclass(GaussianRFConfig, "rng_seed", "rng-seed"),
                ],
            ),
            MutuallyExclusiveArgumentGroup(
                "ms-gaussian",
                title="Multi-scale Gaussian RFs",
                desc="Configuration group for multi-scale orthogonal random fourier features (ORFF), which are used to approximate a Gaussian kernel with multiple length-scales. Specify ``--rf=ms-gaussian`` to enable this RF approximation",
                data_class=MultiscaleGaussianRFConfig,
                arguments=[
                    Argument.from_dataclass(
                        MultiscaleGaussianRFConfig, "num_random_features", "num-rf"
                    ),
                    Argument.from_dataclass(
                        MultiscaleGaussianRFConfig,
                        "use_offset",
                        "use-offset",
                        "no-use-offset",
                    ),
                    Argument.from_dataclass(
                        MultiscaleGaussianRFConfig,
                        "length_scale_low",
                        "length-scale-low",
                    ),
                    Argument.from_dataclass(
                        MultiscaleGaussianRFConfig,
                        "length_scale_high",
                        "length-scale-high",
                    ),
                    Argument.from_dataclass(
                        MultiscaleGaussianRFConfig,
                        "length_scale_num",
                        "length-scale-num",
                    ),
                    Argument.from_dataclass(
                        MultiscaleGaussianRFConfig, "rng_seed", "rng-seed"
                    ),
                ],
            ),
        ],
        help_text="Choose the random-feature approximation.",
    )
    return {
        "solver": solver_arg_group,
        "dataset": dset_arg_group,
        "backbone": bbone_groups,
        "rfs": rf_groups,
    }


def build_parser(return_groups: bool = False):
    parser = argparse.ArgumentParser(
        prog="franken.autotune",
        description="Franken autotune: automatic hyperparameter tuning for franken models",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "--rf-norm",
        choices=["leading_eig", "none"],
        default="leading_eig",
        help=get_field_docstring(AutotuneConfig, "rf_normalization"),
    )
    parser.add_argument(
        "--save-every-model",
        action="store_true",
        help=get_field_docstring(AutotuneConfig, "save_every_model"),
    )
    parser.add_argument(
        "--dtype",
        choices=["float32", "float64"],
        default="float64",
        help=get_field_docstring(AutotuneConfig, "dtype"),
    )
    parser.add_argument(
        "--save-fmaps",
        action="store_true",
        help=get_field_docstring(AutotuneConfig, "save_fmaps"),
    )
    parser.add_argument(
        "--global-scaling",
        action="store_false",
        help=get_field_docstring(AutotuneConfig, "scale_by_species"),
    )
    parser.add_argument(
        "--jac-chunk-size",
        default="auto",
        help=get_field_docstring(AutotuneConfig, "jac_chunk_size"),
    )
    parser.add_argument(
        "--run-dir",
        "-p",
        type=str,
        default=".",
        help=get_field_docstring(AutotuneConfig, "run_dir"),
    )
    parser.add_argument(
        "--seed",
        type=int,
        default=1337,
        help=get_field_docstring(AutotuneConfig, "seed"),
    )
    parser.add_argument(
        "--log-level",
        choices=["DEBUG", "INFO", "WARN", "ERROR"],
        default="INFO",
        help=get_field_docstring(AutotuneConfig, "console_logging_level"),
    )

    arg_groups = get_arg_groups()
    for g in arg_groups.values():
        g.add_to_parser(parser)

    if return_groups:
        return parser, arg_groups
    return parser


def parse_cli(argv):
    parser, groups = build_parser(True)
    args = parser.parse_args(argv)

    groups["backbone"].validate_required(parser, args)
    groups["rfs"].validate_required(parser, args)

    bbone_config = groups["backbone"].to_dataclass(args)
    rf_config = groups["rfs"].to_dataclass(args)
    solver_config = groups["solver"].to_dataclass(args)
    dset_config = groups["dataset"].to_dataclass(args)

    # Initialize autotune config
    autotune_cfg = AutotuneConfig(
        dataset=dset_config,
        solver=solver_config,
        backbone=bbone_config,
        rfs=rf_config,
        rf_normalization=args.rf_norm,
        save_every_model=args.save_every_model,
        dtype=args.dtype,
        save_fmaps=args.save_fmaps,
        scale_by_species=not args.global_scaling,
        jac_chunk_size=args.jac_chunk_size,
        run_dir=args.run_dir,
        seed=args.seed,
        console_logging_level=args.log_level,
    )
    return autotune_cfg



from abc import ABC
import ast
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any, ClassVar, Literal, Sequence, Union


def all_fields(class_or_instance):
    """Return a tuple describing the fields of this dataclass.

    Accepts a dataclass or an instance of one. Tuple elements are of
    type Field.
    """

    try:
        fields = getattr(class_or_instance, "__dataclass_fields__")
    except AttributeError:
        raise TypeError("must be called with a dataclass type or instance") from None

    # Exclude pseudo-fields.  Note that fields is sorted by insertion
    # order, so the order of the tuple is as the fields were defined.
    return tuple(
        f
        for f in fields.values()
        if f._field_type.name == "_FIELD_CLASSVAR" or f._field_type.name == "_FIELD"
    )


def asdict_with_classvar(obj) -> dict[str, Any]:
    if hasattr(obj, "__dataclass_fields__"):
        result = []
        for f in all_fields(obj):
            value = asdict_with_classvar(getattr(obj, f.name))
            result.append((f.name, value))
        return dict(result)
    elif isinstance(obj, tuple) and hasattr(obj, "_fields"):
        return type(obj)(*[asdict_with_classvar(v) for v in obj])
    elif isinstance(obj, (list, tuple)):
        # Assume we can create an object of this type by passing in a
        # generator (which is not true for namedtuples, handled
        # above).
        return type(obj)(asdict_with_classvar(v) for v in obj)
    elif isinstance(obj, dict):
        return type(obj)(
            (asdict_with_classvar(k), asdict_with_classvar(v)) for k, v in obj.items()
        )
    else:
        return deepcopy(obj)


@dataclass
class HPSearchConfig:
    value: float | None = None
    values: Sequence[float] | None = None
    start: float | None = None
    stop: float | None = None
    num: int | None = None
    scale: Literal["log", "linear"] | None = None

    def get_vals(self) -> list[float | int]:
        if self.value is not None:
            return [self.value]
        elif self.values is not None:
            return list(self.values)
        else:
            assert (
                self.start is not None
                and self.stop is not None
                and self.num is not None
                and self.scale is not None
            )
            import numpy as np

            if self.scale == "log":
                return np.logspace(self.start, self.stop, self.num).tolist()
            elif self.scale == "linear":
                return np.linspace(self.start, self.stop, self.num).tolist()
            else:
                raise ValueError(self.scale)

    @staticmethod
    def from_str(hp_str: str) -> "HPSearchConfig":
        try:
            val = float(hp_str)
            return HPSearchConfig(value=val)
        except ValueError:
            pass
        try:
            # Deal with linear, log being unquoted
            hp_str_tmp = [hp_tmp.strip() for hp_tmp in hp_str.strip("( )").split(",")]
            if len(hp_str_tmp) == 4 and hp_str_tmp[-1] in {"log", "linear"}:
                hp_str = f'({hp_str_tmp[0]}, {hp_str_tmp[1]}, {hp_str_tmp[2]}, "{hp_str_tmp[3]}")'
            vals = ast.literal_eval(hp_str)
            if isinstance(vals, (tuple, list)):
                if all(isinstance(v, (int, float)) for v in vals):
                    return HPSearchConfig(values=vals)
                else:
                    if (
                        len(vals) == 4
                        and vals[-1] in {"log", "linear"}
                        and all(isinstance(v, (int, float)) for v in vals[:3])
                    ):
                        return HPSearchConfig(
                            start=vals[0], stop=vals[1], num=vals[2], scale=vals[3]
                        )
            elif isinstance(vals, (float, int)):
                return HPSearchConfig(value=vals)
        except Exception as e:
            raise e
            pass
        raise TypeError(
            f"String '{hp_str}' cannot be converted to an instance of HPSearchConfig. You can either provide float or int values, "
            f"lists of float or int values, or a 4-tuple `(<start>, <stop>, <num>, <'log' | 'linear'>)` which specifies a linear "
            f"or logarithmic range."
        )

    @classmethod
    def is_string_valid(cls, hp_str: str) -> bool:
        try:
            cls.from_str(hp_str)
            return True
        except Exception:
            return False

    def __str__(self) -> str:
        if self.value is not None:
            return f"{self.value}"
        elif self.values is not None:
            return f"{self.values}"
        else:
            return f"({self.start}, {self.stop}, {self.num}, {self.scale})"

    def __post_init__(self):
        if self.value is None:
            if self.values is None:
                if (
                    self.start is None
                    or self.stop is None
                    or self.num is None
                    or self.scale not in {"log", "linear"}
                ):
                    raise TypeError("Invalid HPSearchConfig object was created.")


@dataclass
class DatasetConfig:
    name: str
    """Dataset name. Can be either one of the predefined datasets, or a custom one."""

    train_path: str | None = None
    """Path to the training dataset. This should be readable with `ase`, e.g. xyz files work."""

    test_path: str | None = None
    """Path to the test dataset. When doing hyperparameter tuning this
    is never used to determine the best model."""

    val_path: str | None = None
    """Path to the validation dataset. Used for hyperparameter tuning if available."""

    max_train_samples: int | None = None
    """Randomly subsample the training set to have at most `max_train_samples` points."""


@dataclass
class BackboneConfig(ABC):
    path_or_id: str
    """Path to the GNN checkpoint, or an ID corresponding to a predefined checkpoint.
    To list the available IDs please run the `franken.backbones list` command."""

    interaction_block: int
    """GNN layer out of which the features are extracted."""

    family: ClassVar[str]

    def to_ckpt(self):
        return asdict_with_classvar(self)

    @staticmethod
    def from_ckpt(ckpt):
        if ckpt["family"] == "mace":
            cls = MaceBackboneConfig
        elif ckpt["family"] == "fairchem":
            cls = FairchemBackboneConfig
        elif ckpt["family"] == "sevenn":
            cls = SevennBackboneConfig
        else:
            raise ValueError(ckpt["family"])
        init_args = deepcopy(ckpt)
        init_args.pop("family")
        return cls(**init_args)


@dataclass
class MaceBackboneConfig(BackboneConfig):
    family: ClassVar[str] = "mace"
    interaction_block: int = 2
    """GNN layer out of which the features are extracted."""


@dataclass
class FairchemBackboneConfig(BackboneConfig):
    family: ClassVar[str] = "fairchem"
    interaction_block: int = 2
    """GNN layer out of which the features are extracted."""


@dataclass
class SevennBackboneConfig(BackboneConfig):
    family: ClassVar[str] = "sevenn"
    interaction_block: int = 2
    """GNN layer out of which the features are extracted."""

    extract_after_act: bool = True
    """Whether features should be extracted before or after activations."""

    append_layers: bool = True
    """Whether to take only the features from the last interaction layer, or to concatenate them all."""


@dataclass
class RFConfig(ABC):
    rf_type: ClassVar[str]

    def to_ckpt(self):
        return asdict_with_classvar(self)

    @staticmethod
    def from_ckpt(ckpt):
        if ckpt["rf_type"] == "gaussian":
            cls = GaussianRFConfig
        elif ckpt["rf_type"] == "multiscale-gaussian":
            cls = MultiscaleGaussianRFConfig
        else:
            raise ValueError(ckpt["rf_type"])
        init_args = deepcopy(ckpt)
        init_args.pop("rf_type")
        return cls(**init_args)


@dataclass
class GaussianRFConfig(RFConfig):
    """Orthogonal random features to approximate the Gaussian kernel."""

    rf_type: ClassVar[str] = "gaussian"
    num_random_features: int
    """Number of random features"""

    length_scale: HPSearchConfig | list[float] | float = field(
        default_factory=lambda: HPSearchConfig(start=4, stop=20, num=10, scale="linear")
    )
    """Gaussian length-scale sigma."""

    use_offset: bool = True
    """Whether or not to use the version of Gaussian random features with a uniform offset. Set to true, increases the estimator's efficiency."""

    rng_seed: int = 1337
    """Random seed for reproducibility of the random features"""


@dataclass
class MultiscaleGaussianRFConfig(RFConfig):
    """Random features to approximate a Gaussian kernel with multiple length-scales."""

    rf_type: ClassVar[str] = "multiscale-gaussian"
    num_random_features: int
    """Number of random features"""

    length_scale_low: float = 4
    """Bottom of the range for the length-scale. Should be positive."""

    length_scale_high: float = 20
    """Top of the range for the length-scale. Should be positive."""

    length_scale_num: int = 6
    """Number of different length-scales to include within the specified range."""

    use_offset: bool = True
    """Whether or not to use the version of Gaussian random features with a uniform offset. Set to true, increases the estimator's efficiency."""

    rng_seed: int = 1337
    """Random seed for reproducibility of the random features"""


@dataclass
class SolverConfig:
    l2_penalty: HPSearchConfig | list[float] | float = field(
        default_factory=lambda: HPSearchConfig(start=-11, stop=-6, num=6, scale="log")
    )
    """The amount of regularization. Should be a small positive number."""

    force_weight: HPSearchConfig | list[float] | float = field(
        default_factory=lambda: HPSearchConfig(
            start=0.01, stop=0.99, num=10, scale="linear"
        )
    )
    """Controls how much weight the forces term, as opposed to the energy term has in the loss. Should be a number between 0 and 1."""


@dataclass
class AutotuneConfig:
    """Configure automatic hyperparameter tuning for franken.

    The program will run a grid-search over certain hyperparameters (denoted by 'HYPERPARAMETER' in the help-text), which can be configured by using the following three forms:
    1. a simple float or int value,
    2. a list of float or int values,
    3. a 4-tuple `(<start>, <stop>, <num>, <'log' | 'linear'>)` which specifies a linear or logarithmic range.

    There are two classes of hyperparameters over which we can perform grid-search: solver parameters (for example the L2 penalty), and random feature parameters. It is much more efficient to try multiple combinations of the solver parameters than of the random feature parameters, so plan your grid accordingly.
    In particular, some random feature approximations like the `multiscale-gaussian` do not need any hyperparameter search, reducing the time required for the overall grid-search.

    The program has two sub-commands: `backbone` which allows to choose the GNN underlying the franken features, and `rfs` to choose the random feature approximation and its parameters.
    To view the help-text for the subcommands, you can run for example
    `franken.autotune backbone:mace -h`
    or
    `franken.autotune backbone:mace rfs:gaussian -h`
    Note that the backbone must be chosen before the random features
    """

    dataset: DatasetConfig
    """Configure a dataset for training Franken.

    If `--dataset.name` corresponds to one of the datasets used in the Franken paper (e.g. "water", "PtH2O", "TM23/Ag", etc.) there is no need to specify train, test or validation paths: the code will take care of downloading and preprocessing the data automatically.
    Instead, to use a custom dataset please specify at a minimum the training path, and ideally also the validation path (which is used to determine the best model during a hyperparameter search).
    """

    solver: SolverConfig
    """Configure the franken solver. Hyperparameter search over these is efficient, so the search-grid can be quite fine-grained."""

    backbone: BackboneConfig
    """The GNN backbone which will be used by franken."""

    rfs: RFConfig
    """Choose the random-feature approximation."""

    rf_normalization: Literal["leading_eig"] | None = field(default="leading_eig")
    """Normalization strategy for the covariance matrix."""

    save_every_model: bool = False
    """If true saves a checkpoint for every trial, otherwise it saves only the best model."""

    dtype: Literal["float32", "float64"] = "float64"
    """Data-type for the franken solution. float64 can usually obtain slightly smaller errors while paying a small performance cost."""

    save_fmaps: bool = False
    """Whether to save training feature maps. If the dataset is small (~100 samples), setting this to True can increase the speed of hyperparameter tuning, at the cost of higher memory usage."""

    scale_by_species: bool = True
    """how to scale the GNN features, whether globally (across species) or individually per species."""

    jac_chunk_size: Union[Literal["auto",], int] = "auto"
    """Chunk-size for jacobian calculations. 'auto' attempts to set it based on available GPU memory. If you encounter out-of-memory issues, try setting this manually."""

    run_dir: str = "."
    """Directory in which the hyperparameter search results will be saved"""

    seed: int = 1337
    """Random seed"""

    console_logging_level: Literal["DEBUG", "INFO", "WARN", "ERROR"] = "INFO"
    """Controls verbosity"""



import dataclasses
import datetime
import json
import logging
import sys
import warnings
from pathlib import Path
from typing import Generator, Literal, NamedTuple
from uuid import uuid4

import torch.distributed
import torch.utils.data

from franken.autotune.cli import build_parser, parse_cli
from franken.config import (
    AutotuneConfig,
    BackboneConfig,
    HPSearchConfig,
    RFConfig,
    SolverConfig,
    asdict_with_classvar,
)
from franken.datasets.registry import DATASET_REGISTRY
from franken.trainers.rf_cuda_lowmem import RandomFeaturesTrainer
import franken.utils.distributed as dist_utils
from franken.backbones.utils import CacheDir
from franken.data import BaseAtomsDataset
from franken.rf.model import FrankenPotential
from franken.trainers import BaseTrainer
from franken.trainers.log_utils import DataSplit, LogEntry
from franken.utils.misc import (
    garbage_collection_cuda,
    get_device_name,
    params_grid,
    pprint_config,
    setup_logger,
)


class BestTrial(NamedTuple):
    trial_id: int
    log: LogEntry


warnings.filterwarnings(
    "ignore",
    message=r"You are using `torch.load` with `weights_only=False`",
    category=FutureWarning,
)
warnings.filterwarnings(
    "ignore",
    message=r"`torch.cuda.amp.autocast\(args...\)` is deprecated",
    category=FutureWarning,
)
logger = logging.getLogger("franken")


def init_loaders(
    gnn_cfg: BackboneConfig,
    train_path: Path | None,
    val_path: Path | None = None,
    test_path: Path | None = None,
    num_train_subsamples: int | None = None,
    subsample_rng: int | None = None,
) -> dict[str, torch.utils.data.DataLoader]:
    datasets: dict[str, BaseAtomsDataset] = {}
    for split, data_path in zip(
        ["train", "val", "test"], [train_path, val_path, test_path]
    ):
        if data_path is not None:
            dset = BaseAtomsDataset.from_path(
                data_path=data_path,
                split=split,
                gnn_config=gnn_cfg,
                num_random_subsamples=(
                    num_train_subsamples if split == "train" else None
                ),
                subsample_rng=subsample_rng,
            )
            datasets[split] = dset

    dataloaders = {
        split: dset.get_dataloader(distributed=torch.distributed.is_initialized())
        for split, dset in datasets.items()
    }
    return dataloaders


def hp_summary_str(trial_id: int, current_best: BestTrial, rf_params: RFConfig) -> str:
    hp_summary = f"Trial {trial_id + 1:>3} |"
    for k, v in rf_params.to_ckpt().items():
        fmt_val = format(v, ".3f" if isinstance(v, float) else "")
        hp_summary += f" {k:^7}: {fmt_val:^7} |"
    try:
        energy_error = current_best.log.get_metric("energy_MAE", DataSplit.VALIDATION)
        forces_error = current_best.log.get_metric("forces_MAE", DataSplit.VALIDATION)
    except KeyError:
        energy_error = current_best.log.get_metric("energy_MAE", DataSplit.TRAIN)
        forces_error = current_best.log.get_metric("forces_MAE", DataSplit.TRAIN)

    hp_summary += (
        f" Best trial {current_best.trial_id} (energy {energy_error:.2f} meV/atom - "
        f"forces {forces_error:.1f} meV/Ang)"
    )
    return hp_summary


def hps_from_config(cfg):
    hp_iterators = {}
    for field in dataclasses.fields(cfg):
        hp_def = getattr(cfg, field.name)
        if isinstance(hp_def, HPSearchConfig):
            hp_iterators[field.name] = hp_def.get_vals()
        else:

            hp_iterators[field.name] = [hp_def]
    return hp_iterators


def create_rf_hpsearch_grid(
    cfg: RFConfig,
) -> Generator[tuple[int, RFConfig], None, None]:
    """Convert a random-features configuration object to a sequence of hyper-parameters.

    This function is a thin wrapper over :func:`franken.utils.misc.params_grid` which handles unrolling compact
    hyperparameter specs, such as generating linearly or logarithmically spaced HP values.

    Args:
        cfg: The random features configuration object

    Yields:
        A sequence of simple dictionaries going from hyperparameter name to value.
    """
    hp_iterators = hps_from_config(cfg)
    # Convert grid as dictionary to grid as RFConfig classes
    for exp_id, grid_item in params_grid(hp_iterators):
        yield (exp_id, type(cfg)(**grid_item))


def create_solver_hpsearch_grid(cfg: SolverConfig):
    return hps_from_config(cfg)


def run_autotune(
    gnn_cfg: BackboneConfig,
    rf_cfg: RFConfig,
    solver_cfg: SolverConfig,
    loaders: dict[str, torch.utils.data.DataLoader],
    scale_by_species: bool,
    jac_chunk_size: int | Literal["auto"],
    trainer: BaseTrainer,
):
    current_best = BestTrial(None, None)
    rf_param_grid = create_rf_hpsearch_grid(rf_cfg)
    solver_param_grid = create_solver_hpsearch_grid(solver_cfg)
    for trial_id, rf_params in rf_param_grid:
        logger.debug(f"Autotune iteration with RF parameters {rf_params}")

        assert isinstance(loaders["train"].dataset, BaseAtomsDataset)  # for typing
        model = FrankenPotential(
            gnn_config=gnn_cfg,
            rf_config=rf_params,
            scale_by_Z=scale_by_species,
            num_species=loaders["train"].dataset.num_species,
            atomic_energies=None,
            jac_chunk_size=jac_chunk_size,
        )

        logs, weights = trainer.fit(model, solver_param_grid)
        for split_name, loader in loaders.items():
            logs = trainer.evaluate(
                model,
                loader,
                logs,
                weights,
                metrics=[
                    "energy_MAE",
                    "forces_MAE",
                    "energy_RMSE",
                    "forces_RMSE",
                    "forces_cosim",
                ],
            )
        split_for_best_model = (
            DataSplit.VALIDATION if "val" in loaders else DataSplit.TRAIN
        )
        if dist_utils.get_rank() == 0:
            if trainer.log_dir is not None:
                trainer.serialize_logs(model, logs, weights, split_for_best_model)
        dist_utils.barrier()

        # current best model update
        if dist_utils.get_rank() == 0:
            if trainer.log_dir is not None:
                with open(trainer.log_dir / "best.json", "r") as f:
                    try:
                        best_log = LogEntry.from_dict(json.load(f))
                        if best_log != current_best.log:
                            current_best = BestTrial(
                                trial_id=trial_id + 1,
                                log=best_log,
                            )
                    except KeyError:
                        pass

                logger.info(hp_summary_str(trial_id, current_best, rf_params))
        garbage_collection_cuda()


def create_run_folder(base_run_dir: Path) -> Path:
    # Use time + a shortened UUID to ensure uniqueness of
    # the experiment directory. The names will look like
    # 'run_240926_113513_1d93b3ed'
    now_str = datetime.datetime.now().strftime("%y%m%d_%H%M%S")
    exp_dir_name = f"run_{now_str}_{uuid4().hex[:8]}"
    exp_dir = base_run_dir / exp_dir_name
    exp_dir.mkdir(parents=True, exist_ok=True)
    return exp_dir


def save_experiment_info(cfg: AutotuneConfig, run_dir: Path):
    train_hardware = {
        "num_gpus": dist_utils.get_world_size(),
        "gpu_model": get_device_name("cuda:0"),
        "cpu_model": get_device_name("cpu"),
    }
    cfg_dict = asdict_with_classvar(cfg)
    assert isinstance(cfg_dict, dict)
    with open(run_dir / "configs.json", "w") as f:
        json.dump(cfg_dict | train_hardware, f, indent=4)


def get_dataset_paths(
    train_path: str | None,
    val_path: str | None,
    test_path: str | None,
    dataset_name: str | None,
) -> tuple[Path, Path | None, Path | None]:
    try:
        if dataset_name is None:
            raise KeyError
        out_train_path = DATASET_REGISTRY.get_path(
            dataset_name, "train", CacheDir.get()
        )
        out_val_path = None
        if DATASET_REGISTRY.is_valid_split(dataset_name, "val"):
            out_val_path = DATASET_REGISTRY.get_path(
                dataset_name, "val", CacheDir.get()
            )
        out_test_path = None
        if DATASET_REGISTRY.is_valid_split(dataset_name, "test"):
            out_test_path = DATASET_REGISTRY.get_path(
                dataset_name, "test", CacheDir.get()
            )
        if out_val_path is not None and out_test_path is not None:
            out_test_path = (
                None  # TODO: This is not very good, check with our datasets!
            )
    except KeyError:
        logger.info(f"Dataset with name '{dataset_name}' not found in registry")
        if train_path is None:
            raise ValueError(
                "Either a valid 'dataset_name' or 'train_path' must be "
                "specified in order to load a training dataset."
            )
        out_train_path = Path(train_path)
        out_val_path = Path(val_path) if val_path is not None else None
        out_test_path = Path(test_path) if test_path is not None else None
    return out_train_path, out_val_path, out_test_path


def autotune(cfg: AutotuneConfig):
    run_dir = Path(cfg.run_dir)

    if torch.cuda.is_available():
        rank = dist_utils.init(distributed=torch.cuda.device_count() > 1)
        device = torch.device(torch.cuda.current_device())
    else:
        rank = 0
        device = torch.device("cpu")

    if rank != 0:  # first rank goes forward
        run_dir = None
        dist_utils.barrier()
    else:
        run_dir = create_run_folder(run_dir)
        dist_utils.barrier()  # other ranks follow
    logging_level = cfg.console_logging_level.upper()
    setup_logger(
        level=logging_level, directory=dist_utils.broadcast_obj(run_dir), rank=rank
    )
    pprint_config(asdict_with_classvar(cfg))

    # Global try-catch after setup_logger, to log any exceptions.
    try:
        CacheDir.initialize()

        if rank != 0:  # first rank goes forward
            dist_utils.barrier()
        else:
            assert run_dir is not None
            save_experiment_info(cfg, run_dir)
            logger.info(f"Run folder: {run_dir}")
            dist_utils.barrier()

        train_path, val_path, test_path = get_dataset_paths(
            cfg.dataset.train_path,
            cfg.dataset.val_path,
            cfg.dataset.test_path,
            cfg.dataset.name,
        )

        loaders = init_loaders(
            cfg.backbone,
            train_path,
            val_path,
            test_path,
            cfg.dataset.max_train_samples,
            cfg.seed,
        )

        trainer = RandomFeaturesTrainer(
            train_dataloader=loaders["train"],
            random_features_normalization=cfg.rf_normalization,
            save_every_model=cfg.save_every_model,
            dtype=cfg.dtype,
            save_fmaps=cfg.save_fmaps,
            log_dir=run_dir,
            device=device,
        )

        run_autotune(
            gnn_cfg=cfg.backbone,
            rf_cfg=cfg.rfs,
            solver_cfg=cfg.solver,
            loaders=loaders,
            scale_by_species=cfg.scale_by_species,
            jac_chunk_size=cfg.jac_chunk_size,
            trainer=trainer,
        )
    except Exception as e:
        logger.error("Error encountered in autotune. Exiting.", exc_info=e)
        raise
    finally:
        if torch.distributed.is_initialized():
            torch.distributed.destroy_process_group()

    return run_dir


def cli_entry_point():
    args = parse_cli(sys.argv[1:])
    autotune(args)


if __name__ == "__main__":
    cli_entry_point()


# For sphinx docs
get_parser_fn = build_parser()
